package handler

import (
	"bytes"
	"encoding/json"
	"fmt"
	"gosimplenpm/internal/config"
	"gosimplenpm/internal/serviceidos"
	"gosimplenpm/internal/storage"
	"io"
	"net/http"
	"net/http/httptest"
	"net/url"
	"os"
	"testing"

	"github.com/gorilla/mux"
	"github.com/sirupsen/logrus"
	"github.com/stretchr/testify/assert"
)

func TestUnitTagGet(t *testing.T) {
	t.Run("return `Internal Server` error if index json cannot be retrieved", func(t *testing.T) {
		req := httptest.NewRequest(http.MethodGet, "/-/package/{name}/dist-tags", nil)
		wrt := httptest.NewRecorder()

		lg := &logrus.Logger{
			Out: os.Stdout,
			// Level: "DEBUG",
			Formatter: &logrus.TextFormatter{
				FullTimestamp:   true,
				TimestampFormat: "2009-01-02 15:15:15",
			},
		}

		cfg := config.Config{
			RepoDir: "",
		}

		mfs := &storage.MockFs{}
		mfs.GetIndexJsonFromStoreFunc = func(packageName string, registryPath string, lg *logrus.Logger) (string, bool, error) {
			return "", false, fmt.Errorf("Filesystem error")
		}

		tag := "1.2.4"

		//Hack to try to fake gorilla/mux vars
		vars := map[string]string{
			"name": "test-package",
			"tag":  tag,
		}

		req = mux.SetURLVars(req, vars)

		DistTagGet(lg, cfg, mfs)(wrt, req)

		rs := wrt.Result()

		assert.Equal(t, rs.StatusCode, http.StatusInternalServerError)

		defer rs.Body.Close()
		body, err := io.ReadAll(rs.Body)
		if err != nil {
			t.Fatal(err)
		}
		bytes.TrimSpace(body)

		assert.Equal(t, string(body), "Filesystem error\n")
	})

	t.Run("return `Not Found` error if package cannot be retrieved", func(t *testing.T) {
		req := httptest.NewRequest(http.MethodGet, "/-/package/{name}/dist-tags", nil)
		wrt := httptest.NewRecorder()

		lg := &logrus.Logger{
			Out: os.Stdout,
			// Level: "DEBUG",
			Formatter: &logrus.TextFormatter{
				FullTimestamp:   true,
				TimestampFormat: "2009-01-02 15:15:15",
			},
		}

		cfg := config.Config{
			RepoDir: "",
		}

		mfs := &storage.MockFs{}
		mfs.GetIndexJsonFromStoreFunc = func(packageName string, registryPath string, lg *logrus.Logger) (string, bool, error) {
			return "", false, nil
		}

		//Hack to try to fake gorilla/mux vars
		vars := map[string]string{
			"name": "test-package",
		}

		req = mux.SetURLVars(req, vars)

		DistTagGet(lg, cfg, mfs)(wrt, req)

		rs := wrt.Result()

		assert.Equal(t, rs.StatusCode, http.StatusNotFound)

		defer rs.Body.Close()
		body, err := io.ReadAll(rs.Body)
		if err != nil {
			t.Fatal(err)
		}
		bytes.TrimSpace(body)

		assert.Equal(t, string(body), "Package not found: test-package\n")
	})

	t.Run("return `Internal Server` error if package.json cannot be parsed", func(t *testing.T) {
		req := httptest.NewRequest(http.MethodDelete, "/-/package/{name}/dist-tags", nil)
		wrt := httptest.NewRecorder()

		lg := &logrus.Logger{
			Out: os.Stdout,
			// Level: "DEBUG",
			Formatter: &logrus.TextFormatter{
				FullTimestamp:   true,
				TimestampFormat: "2009-01-02 15:15:15",
			},
		}

		cfg := config.Config{
			RepoDir: "",
		}

		mfs := &storage.MockFs{}
		mfs.GetIndexJsonFromStoreFunc = func(packageName string, registryPath string, lg *logrus.Logger) (string, bool, error) {
			return "", true, nil
		}
		mfs.ReadIndexJsonFunc = func(fPath string, res *serviceidos.IndexJson, lg *logrus.Logger) error {
			return fmt.Errorf("Parsing failed")
		}

		//Hack to try to fake gorilla/mux vars
		vars := map[string]string{
			"name": "test-package",
		}

		req = mux.SetURLVars(req, vars)

		DistTagGet(lg, cfg, mfs)(wrt, req)

		rs := wrt.Result()

		assert.Equal(t, rs.StatusCode, http.StatusInternalServerError)

		defer rs.Body.Close()
		body, err := io.ReadAll(rs.Body)
		if err != nil {
			t.Fatal(err)
		}
		bytes.TrimSpace(body)

		assert.Equal(t, string(body), "Parsing failed\n")
	})

	t.Run("return 200 OK", func(t *testing.T) {
		req := httptest.NewRequest(http.MethodDelete, "/-/package/{name}/dist-tags", nil)
		wrt := httptest.NewRecorder()

		lg := &logrus.Logger{
			Out: os.Stdout,
			// Level: "DEBUG",
			Formatter: &logrus.TextFormatter{
				FullTimestamp:   true,
				TimestampFormat: "2009-01-02 15:15:15",
			},
		}

		cfg := config.Config{
			RepoDir: "",
		}

		mfs := &storage.MockFs{}
		mfs.GetIndexJsonFromStoreFunc = func(packageName string, registryPath string, lg *logrus.Logger) (string, bool, error) {
			return "", true, nil
		}
		mfs.ReadIndexJsonFunc = func(fPath string, res *serviceidos.IndexJson, lg *logrus.Logger) error {
			temp := serviceidos.IndexJson{
				DistTags: map[string]string{
					"current": "1.2.0",
					"latest":  "1.2.0",
				},
			}
			*res = temp
			return nil
		}

		//Hack to try to fake gorilla/mux vars
		vars := map[string]string{
			"name": "test-package",
		}

		req = mux.SetURLVars(req, vars)

		DistTagGet(lg, cfg, mfs)(wrt, req)

		rs := wrt.Result()

		assert.Equal(t, rs.StatusCode, http.StatusOK)

		defer rs.Body.Close()
		body, err := io.ReadAll(rs.Body)
		if err != nil {
			t.Fatal(err)
		}
		bytes.TrimSpace(body)

		assert.Equal(t, string(body), "{\"current\":\"1.2.0\",\"latest\":\"1.2.0\"}")
	})
}

func TestIntegrationTagGet(t *testing.T) {
	if testing.Short() {
		t.Skip("Skipping getTags integration test")
	}

	token := "0N89nr/hmKXoBzG]R{fKH%YE1X"

	tmpDir := t.TempDir()

	mkDir(t, fmt.Sprintf("%s/@df/simplepackone", tmpDir))
	mkDir(t, fmt.Sprintf("%s/output", tmpDir))

	indexJsonFp := "intestdata/tags/index.json"
	tgzFp := "intestdata/tags/simplepackone-1.9.0.tgz"

	cpFile(t, indexJsonFp, fmt.Sprintf("%s/@df/simplepackone/index.json", tmpDir))
	cpFile(t, tgzFp, fmt.Sprintf("%s/@df/simplepackone/simplepackone-1.9.0.tgz", tmpDir))

	cfg := config.Config{
		RepoDir: tmpDir,
		Token:   token,
	}

	app := newTestApp(t, cfg)
	app.Routes()
	ts := newTestServer(t, app.Mux)
	defer ts.Close()

	code, _, body := ts.get(t, fmt.Sprintf("/-/package/%s/dist-tags", url.PathEscape("@df/simplepackone")))

	assert.Equal(t, code, http.StatusOK)
	ret := make(map[string]interface{})
	json.Unmarshal(body, &ret)
	assert.Contains(t, ret, "latest")
	assert.Equal(t, "1.9.0", ret["latest"])
}
